from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generic,
    Literal,
    Optional,
    Tuple,
    Type,
)

from bluesky.protocols import DataKey, Reading

from .utils import DEFAULT_TIMEOUT, ReadingValueCallback, T


class SignalBackend(Generic[T]):
    """A read/write/monitor backend for a Signals"""

    #: Datatype of the signal value
    datatype: Optional[Type[T]] = None

    #: Like ca://PV_PREFIX:SIGNAL
    @abstractmethod
    def source(self, name: str) -> str:
        """Return source of signal. Signals may pass a name to the backend, which can be
        used or discarded."""

    @abstractmethod
    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        """Connect to underlying hardware"""

    @abstractmethod
    async def put(self, value: Optional[T], wait=True, timeout=None):
        """Put a value to the PV, if wait then wait for completion for up to timeout"""

    @abstractmethod
    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""

    @abstractmethod
    async def get_reading(self) -> Reading:
        """The current value, timestamp and severity"""

    @abstractmethod
    async def get_value(self) -> T:
        """The current value"""

    @abstractmethod
    async def get_setpoint(self) -> T:
        """The point that a signal was requested to move to."""

    @abstractmethod
    def set_callback(self, callback: Optional[ReadingValueCallback[T]]) -> None:
        """Observe changes to the current value, timestamp and severity"""


class _RuntimeSubsetEnumMeta(type):
    def __str__(cls):
        if hasattr(cls, "choices"):
            return f"SubsetEnum{list(cls.choices)}"
        return "SubsetEnum"

    def __getitem__(cls, _choices):
        if isinstance(_choices, str):
            _choices = (_choices,)
        else:
            if not isinstance(_choices, tuple) or not all(
                isinstance(c, str) for c in _choices
            ):
                raise TypeError(
                    "Choices must be a str or a tuple of str, " f"not {type(_choices)}."
                )
            if len(set(_choices)) != len(_choices):
                raise TypeError("Duplicate elements in runtime enum choices.")

        class _RuntimeSubsetEnum(cls):
            choices = _choices

        return _RuntimeSubsetEnum


class RuntimeSubsetEnum(metaclass=_RuntimeSubsetEnumMeta):
    choices: ClassVar[Tuple[str, ...]]

    def __init__(self):
        raise RuntimeError("SubsetEnum cannot be instantiated")


if TYPE_CHECKING:
    SubsetEnum = Literal
else:
    SubsetEnum = RuntimeSubsetEnum



from unittest.mock import patch

import pytest

from ophyd_async.core import DetectorTrigger, DeviceCollector, set_mock_value
from ophyd_async.epics.areadetector.controllers import (
    ADSimController,
    PilatusController,
)
from ophyd_async.epics.areadetector.drivers import ADBase, PilatusDriver
from ophyd_async.epics.areadetector.drivers.pilatus_driver import PilatusTriggerMode
from ophyd_async.epics.areadetector.utils import ImageMode


@pytest.fixture
async def pilatus_driver(RE) -> PilatusDriver:
    async with DeviceCollector(mock=True):
        drv = PilatusDriver("DRIVER:")

    return drv


@pytest.fixture
async def pilatus(RE, pilatus_driver: PilatusDriver) -> PilatusController:
    async with DeviceCollector(mock=True):
        controller = PilatusController(pilatus_driver, readout_time=2.28)

    return controller


@pytest.fixture
async def ad(RE) -> ADSimController:
    async with DeviceCollector(mock=True):
        drv = ADBase("DRIVER:")
        controller = ADSimController(drv)

    return controller


async def test_ad_controller(RE, ad: ADSimController):
    with patch("ophyd_async.core.signal.wait_for_value", return_value=None):
        await ad.arm(num=1)

    driver = ad.driver
    assert await driver.num_images.get_value() == 1
    assert await driver.image_mode.get_value() == ImageMode.multiple
    assert await driver.acquire.get_value() is True

    with patch(
        "ophyd_async.epics.areadetector.utils.wait_for_value", return_value=None
    ):
        await ad.disarm()

    assert await driver.acquire.get_value() is False


async def test_pilatus_controller(
    RE,
    pilatus: PilatusController,
    pilatus_driver: PilatusDriver,
):
    set_mock_value(pilatus_driver.armed_for_triggers, True)
    status = await pilatus.arm(num=1, trigger=DetectorTrigger.constant_gate)
    await status

    assert await pilatus_driver.num_images.get_value() == 1
    assert await pilatus_driver.image_mode.get_value() == ImageMode.multiple
    assert (
        await pilatus_driver.trigger_mode.get_value() == PilatusTriggerMode.ext_enable
    )
    assert await pilatus_driver.acquire.get_value() is True

    await pilatus.disarm()

    assert await pilatus_driver.acquire.get_value() is False



import asyncio
from pathlib import Path
from typing import AsyncGenerator, AsyncIterator, Dict, List, Optional

from bluesky.protocols import DataKey, StreamAsset
from p4p.client.thread import Context

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorWriter,
    DirectoryProvider,
    wait_for_value,
)
from ophyd_async.core.signal import observe_value

from .._common_blocks import CommonPandaBlocks
from ._panda_hdf_file import _HDFDataset, _HDFFile


class PandaHDFWriter(DetectorWriter):
    _ctxt: Optional[Context] = None

    def __init__(
        self,
        directory_provider: DirectoryProvider,
        panda_device: CommonPandaBlocks,
    ) -> None:
        self.panda_device = panda_device
        self._directory_provider = directory_provider
        self._datasets: List[_HDFDataset] = []
        self._file: Optional[_HDFFile] = None
        self._multiplier = 1

    # Triggered on PCAP arm
    async def open(self, multiplier: int = 1) -> Dict[str, DataKey]:
        """Retrieve and get descriptor of all PandA signals marked for capture"""

        # Ensure flushes are immediate
        await self.panda_device.data.flush_period.set(0)

        self._file = None
        info = self._directory_provider()
        # Set the initial values
        await asyncio.gather(
            self.panda_device.data.hdf_directory.set(
                str(info.root / info.resource_dir)
            ),
            self.panda_device.data.hdf_file_name.set(
                f"{info.prefix}{self.panda_device.name}{info.suffix}.h5",
            ),
            self.panda_device.data.num_capture.set(0),
        )

        # Wait for it to start, stashing the status that tells us when it finishes
        await self.panda_device.data.capture.set(True)
        if multiplier > 1:
            raise ValueError(
                "All PandA datasets should be scalar, multiplier should be 1"
            )

        return await self._describe()

    async def _describe(self) -> Dict[str, DataKey]:
        """
        Return a describe based on the datasets PV
        """

        await self._update_datasets()
        describe = {
            ds.data_key: DataKey(
                source=self.panda_device.data.hdf_directory.source,
                shape=ds.shape,
                dtype="array" if ds.shape != [1] else "number",
                external="STREAM:",
            )
            for ds in self._datasets
        }
        return describe

    async def _update_datasets(self) -> None:
        """
        Load data from the datasets PV on the panda, update internal
        representation of datasets that the panda will write.
        """

        capture_table = await self.panda_device.data.datasets.get_value()
        self._datasets = [
            _HDFDataset(dataset_name, "/" + dataset_name, [1], multiplier=1)
            for dataset_name in capture_table["name"]
        ]

    # Next few functions are exactly the same as AD writer. Could move as default
    # StandardDetector behavior
    async def wait_for_index(
        self, index: int, timeout: Optional[float] = DEFAULT_TIMEOUT
    ):
        def matcher(value: int) -> bool:
            return value >= index

        matcher.__name__ = f"index_at_least_{index}"
        await wait_for_value(
            self.panda_device.data.num_captured, matcher, timeout=timeout
        )

    async def get_indices_written(self) -> int:
        return await self.panda_device.data.num_captured.get_value()

    async def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        """Wait until a specific index is ready to be collected"""
        async for num_captured in observe_value(
            self.panda_device.data.num_captured, timeout
        ):
            yield num_captured // self._multiplier

    async def collect_stream_docs(
        self, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        # TODO: fail if we get dropped frames
        if indices_written:
            if not self._file:
                self._file = _HDFFile(
                    self._directory_provider(),
                    Path(await self.panda_device.data.hdf_file_name.get_value()),
                    self._datasets,
                )
                for doc in self._file.stream_resources():
                    yield "stream_resource", doc
            for doc in self._file.stream_data(indices_written):
                yield "stream_datum", doc

    # Could put this function as default for StandardDetector
    async def close(self):
        await self.panda_device.data.capture.set(
            False, wait=True, timeout=DEFAULT_TIMEOUT
        )



import asyncio
from unittest.mock import Mock

import pytest

from ophyd_async.core import (
    DetectorControl,
    DeviceCollector,
    get_mock_put,
    set_mock_value,
)
from ophyd_async.epics.areadetector.drivers import (
    ADBase,
    DetectorState,
    set_exposure_time_and_acquire_period_if_supplied,
    start_acquiring_driver_and_ensure_status,
)

TEST_DEADTIME = 0.1


@pytest.fixture
def driver(RE) -> ADBase:
    with DeviceCollector(mock=True):
        driver = ADBase("DRV:", name="drv")
    return driver


@pytest.fixture
async def controller(RE, driver: ADBase) -> Mock:
    controller = Mock(spec=DetectorControl)
    controller.get_deadtime.return_value = TEST_DEADTIME
    return controller


async def test_set_exposure_time_and_acquire_period_if_supplied_is_a_noop_if_no_exposure_supplied(  # noqa: E501
    controller: DetectorControl,
    driver: ADBase,
):
    put_exposure = get_mock_put(driver.acquire_time)
    put_acquire_period = get_mock_put(driver.acquire_period)
    await set_exposure_time_and_acquire_period_if_supplied(controller, driver, None)
    put_exposure.assert_not_called()
    put_acquire_period.assert_not_called()


@pytest.mark.parametrize(
    "exposure,expected_exposure,expected_acquire_period",
    [
        (0.0, 0.0, 0.1),
        (1.0, 1.0, 1.1),
        (1.5, 1.5, 1.6),
    ],
)
async def test_set_exposure_time_and_acquire_period_if_supplied_uses_deadtime(
    controller: DetectorControl,
    driver: ADBase,
    exposure: float,
    expected_exposure: float,
    expected_acquire_period: float,
):
    await set_exposure_time_and_acquire_period_if_supplied(controller, driver, exposure)
    actual_exposure = await driver.acquire_time.get_value()
    actual_acquire_period = await driver.acquire_period.get_value()
    assert expected_exposure == actual_exposure
    assert expected_acquire_period == actual_acquire_period


async def test_start_acquiring_driver_and_ensure_status_flags_immediate_failure(
    driver: ADBase,
):
    set_mock_value(driver.detector_state, DetectorState.Error)
    acquiring = await start_acquiring_driver_and_ensure_status(driver, timeout=0.01)
    with pytest.raises(ValueError):
        await acquiring


async def test_start_acquiring_driver_and_ensure_status_fails_after_some_time(
    driver: ADBase,
):
    """This test ensures a failing status is captured halfway through acquisition.

    Real world application; it takes some time to start acquiring, and during that time
    the detector gets itself into a bad state.
    """
    set_mock_value(driver.detector_state, DetectorState.Idle)

    async def wait_then_fail():
        await asyncio.sleep(0)
        set_mock_value(driver.detector_state, DetectorState.Disconnected)

    acquiring = await start_acquiring_driver_and_ensure_status(driver, timeout=0.1)
    await wait_then_fail()

    with pytest.raises(ValueError):
        await acquiring



import pytest
from epicscorelibs.ca import dbr
from p4p import Value as P4PValue
from p4p.nt import NTEnum

from ophyd_async.core import SubsetEnum
from ophyd_async.epics._backend._aioca import make_converter as aioca_make_converter
from ophyd_async.epics._backend._p4p import make_converter as p4p_make_converter
from ophyd_async.epics.signal.signal import epics_signal_rw


async def test_runtime_enum_behaviour():
    rt_enum = SubsetEnum["A", "B"]

    with pytest.raises(RuntimeError) as exc:
        rt_enum()
    assert str(exc.value) == "SubsetEnum cannot be instantiated"

    assert issubclass(rt_enum, SubsetEnum)

    # Our metaclass doesn't cache already created runtime enums,
    # so we can't do this
    assert not issubclass(rt_enum, SubsetEnum["A", "B"])
    assert not issubclass(rt_enum, SubsetEnum["B", "A"])
    assert rt_enum is not SubsetEnum["A", "B"]
    assert rt_enum is not SubsetEnum["B", "A"]

    assert str(rt_enum) == "SubsetEnum['A', 'B']"
    assert str(SubsetEnum) == "SubsetEnum"

    with pytest.raises(TypeError) as exc:
        SubsetEnum["A", "B", "A"]
    assert str(exc.value) == "Duplicate elements in runtime enum choices."


async def test_ca_runtime_enum_converter():
    class EpicsValue:
        def __init__(self):
            self.name = "test"
            self.ok = (True,)
            self.errorcode = 0
            self.datatype = dbr.DBR_ENUM
            self.element_count = 1
            self.severity = 0
            self.status = 0
            self.raw_stamp = (0,)
            self.timestamp = 0
            self.datetime = 0
            self.enums = ["A", "B", "C"]  # More than the runtime enum

    epics_value = EpicsValue()
    rt_enum = SubsetEnum["A", "B"]
    converter = aioca_make_converter(
        rt_enum, values={"READ_PV": epics_value, "WRITE_PV": epics_value}
    )
    assert converter.choices == {"A": "A", "B": "B", "C": "C"}
    assert set(rt_enum.choices).issubset(set(converter.choices.keys()))


async def test_pva_runtime_enum_converter():
    enum_type = NTEnum.buildType()
    epics_value = P4PValue(
        enum_type,
        {
            "value.choices": ["A", "B", "C"],
        },
    )
    rt_enum = SubsetEnum["A", "B"]
    converter = p4p_make_converter(
        rt_enum, values={"READ_PV": epics_value, "WRITE_PV": epics_value}
    )
    assert {"A", "B"}.issubset(set(converter.choices))


async def test_runtime_enum_signal():
    signal_rw_pva = epics_signal_rw(SubsetEnum["A1", "B1"], "ca://RW_PV", name="signal")
    signal_rw_ca = epics_signal_rw(SubsetEnum["A2", "B2"], "ca://RW_PV", name="signal")
    await signal_rw_pva.connect(mock=True)
    await signal_rw_ca.connect(mock=True)
    await signal_rw_pva.get_value() == "A1"
    await signal_rw_ca.get_value() == "A2"
    await signal_rw_pva.set("B1")
    await signal_rw_ca.set("B2")
    await signal_rw_pva.get_value() == "B1"
    await signal_rw_ca.get_value() == "B2"

    # Will accept string values even if they're not in the runtime enum
    # Though type checking should compain
    await signal_rw_pva.set("C1")  # type: ignore
    await signal_rw_ca.set("C2")  # type: ignore
